#ifndef cppm_out_h
#define cppm_out_h

#include <Arduino.h>

// driving 6-channel cppm signal on pin 3 (PD4) and pin 13 (PB5) using Timer 2 interrupts - ceptimus 2015-06-25
// Despite articles I found on the internet saying otherwise, my FF6 gives out a buddy box signal
// that is high (5V) most of the time with low (0V) pulses 300 uS wide.  As far as I can tell using
// an oscilloscope, the signal to the RF stage is the same polarity, but at a smaller voltage.
// The Taranis can output a mostly high or mostly low PPM signal to the JR RF module depending on 
// the PPM frame setting in the model menu (first page). The buddy box signal can also be set to +
// or - once it's switched on by selecting 'slave' mode on the model first page.
// The Turnigy 9XR buddy signal (JR port) is only present when the transmitter power switch is off
// it's affected by the +/- setting of PPM frame.  On the Turnigy, the RF signal and the Buddy signal
// always have the same polarity as each other (as chosen by the PPM frame selection).
// For a reference that can't be altered, the JR X-388S has a buddy output mostly high with low pulses
// not much of a voltage swing - maybe open collector.  The output to the RF module is also mostly high
// with low pulses - much bigger swing of the full battery voltage (9.7V when testing)
// on testing my transmitter with a (Futabla format) SPEKTRUM DM8 module, I found it wouldn't work with
// mostly high with low pulses signal, but was happy with the mostly low, high pulses format.
// So conclusion for Futaba format: RF PPM signal mostly low, high pulses
//                       JR format: RF PPM signal mostly high, low pulses
//                                  both have buddy box signals that are mostly high with low pulses.

// The JR module pinout is:
// Top.    PPM signal to module, as described above *
//         Some internet sources say this is +6V - but not on the Turnigy.  I suspect that it may have been an RF-power-level indication signal on old non-UHF modules
//         Battery +V (11.8V on Turnigy)*
//         GND *
// Bottom. RF to Antenna (mostly hangover from non-UHF systems)

// The Futaba module pinout is:
// Nearest to corner.      PPM signal to module, as described above *
//                         Battery  +V. *
//                         RF power output signal to meter on transmitter
//                         GND *
// Furthest from corner.   RF to Antenna (mostly hangover from non-UHF systems)
//

// Only three pins marked * need be connected for operation with Arduino transmitter

class cppm_out {
  public:
    // pulse intervals in half-microsecond units - varies approx 1600~4400 with centre at ~3000 but with 600 subtracted for 300 uS low pulse
    // don't set these directly - they are public as they are accessed from the (non class member) interrupt service routine
    // but to set the 16-bit quantity in an interrupt safe way, use the set() method
    static volatile uint16_t pulseWidth[6]; // pulse widths in half-microseconds

    static void start(void); // call from setup() to initialise output pin, interrupts etc and begin outputting signal
    static void set(uint8_t n, int16_t p); // set channel[n] to position p. n is channel 0 to 5. p ~ +/- 1400
    static void set_uS(uint8_t n, int16_t uS); // set channel[n] pulse time to uS
    static int16_t get_uS(uint8_t n); // return current channel[n] pulse time in uS
};
#endif
