#include "encoder.h"

uint8_t encoder::startupSynch;
uint8_t encoder::state;
int16_t encoder::count;
int16_t encoder::debounce;
int16_t encoder::timeout;
int16_t encoder::multiplier;
int16_t encoder::turningTime;
boolean encoder::pressed;
boolean encoder::multiplying;
boolean encoder::clockwise;
boolean encoder::oldClockwise;

ISR(PCINT1_vect) { // pinchange occured on A0 or A2
  // least significant 4 bits in state are oldA2,A2,oldA0,A0 
  encoder::state = ((encoder::state & 0x03) << 2) | (PINC & 0x03);
  switch(encoder::state) {
    case B0001: // +ve transition A0 while A1 low
      encoder::count--;
      break;
    case B0010: // +ve transition on A1 while A0 low
      encoder::count++;
      break;
    case B0100: // -ve transition on A0 while A1 low
      encoder::count++;
      break;
    case B0111: // +ve transition A1 while A0 high
      encoder::count--;
      break;
    case B1000: // -ve transition A1 while A0 low
      encoder::count--;
      break;
    case B1011: // +ve transition on A0 while A1 high
      encoder::count++;
      break;
    case B1101: // -ve transition on A1 while A0 high
      encoder::count++;
      break;
    case B1110: // -ve transition on A0 while A1 high
      encoder::count--;
      break;
  }
  encoder::timeout = 30000; // reset timeout when encoder turned
}

void encoder::start(void) {
    pinMode(10, INPUT_PULLUP);
    pinMode(A0, INPUT_PULLUP);
    pinMode(A1, INPUT_PULLUP);
    count = 0;
    state = 0;
    multiplier = 1;
    multiplying = false;
    turningTime = 0;
    pressed = false;
    debounce = 0;
    PCMSK1 = 0x03; // mask bits of PORTC (A0 - A7) so that only A0 and A1 generate pinchange interrupts
    PCICR |= 0x02; // enable pinchange interrupts for PORTC
    startupSynch = 0;
}

void encoder::checkEncoder(uint16_t mS) { // mS is milliseconds since previous call
  static boolean oldButton = false; // true if previously pressed
  boolean button; // true when pressed (debounced)

  timeout -= mS;
  if (timeout < 0) {
    timeout = 0;
  }

  if (timeout < 29000) { // switch off multiplying after one second of non-turning
    multiplying = false;
    turningTime = 0;
    if (startupSynch == 1) { // after first detected turn, followed by a second of inactivity
      cli(); count = 0; sei(); // set count to zero to remove any hysteresis deadband. 
      startupSynch = 2; //
    }
  } else { // switch on multiplying after four seconds of turning
    if (turningTime < 4000) {
      cli();
      turningTime += mS;
      sei();
    } else {
      multiplying = true;
    }
  }


  if (!digitalRead(10)) { // undebounced press
    button = true; // comes on at first contact - debouncing is on release
    debounce = 100;
  } else if (debounce > 0) {
    button = true;
    debounce -=mS;
  } else {
    button = false;
  }
  pressed = button && !oldButton; // pressed is true for one scan on first -ve edge of D10 after D10 has been high for 100mS
  oldButton = button;
}

void encoder::setMultiplier(int16_t ratio) {
  if (multiplier != ratio) {
    multiplier = ratio;
    multiplying = false;
  }
}

int16_t encoder::change(int16_t value) {
  int16_t delta;
  cli(); delta = count; sei();
  delta /= ENCODER_COUNTS_PER_CLICK;
  if (delta) {
    clockwise = delta > 0;
    if (clockwise != oldClockwise) {
      multiplying = false;
      turningTime = 0;
      oldClockwise = clockwise;
    }
    if (multiplying)
      value += delta * multiplier;
    else
      value += delta;
      
    delta *= ENCODER_COUNTS_PER_CLICK;
    cli(); count -= delta; sei();
    if (!startupSynch) startupSynch = 1;
  }
  return value;
}

