#include "joystick.h"
#include "eeprom.h"

int16_t joystick::axis[2]; // axis[0] is x (right +ve) scaled to -512 0 512.  axis[1] is y (down (i.e. up elevator) +ve
int16_t joystick::autoRepeat;
int16_t joystick::debounce = 0;
int16_t joystick::buttonDebounce[4]; // for debouncing the left, top, right, bottom push buttons, if fitted
boolean joystick::digital; // digital is true for button joystick, false for analogue version
int16_t joystick::timeout;
uint8_t joystick::button; // bits 0, 1, 2, 3, 4 = right, left, down, up, pressed
uint8_t joystick::increment; // starts at 1 and increases when auto-repeating

void joystick::start(boolean buttonJoystick) { // buttonJoystick is true for button joystick, false for analogue version
  digital = buttonJoystick;
  pinMode(A0, digital ? INPUT_PULLUP : INPUT); // reads x-axis pot or 'left' button
  pinMode(A1, digital ? INPUT_PULLUP : INPUT); // reads y-axis pot or 'top' button
  pinMode(A3, INPUT_PULLUP); // reads 'right' button for button joystick option
  pinMode(A4, INPUT_PULLUP); // reads 'bottom' button for button joystick option
  pinMode(10, INPUT_PULLUP); // for pressed digitial input - switch on internal 'pull up'
}

void joystick::readJoy(uint16_t mS) { // mS is milliseconds since previous call
  static uint8_t oldButton = 0;

  timeout -= mS;
  if (timeout < 0)
    timeout = 0;

  if (digital) { // button joystick option
    axis[0] = axis[1] = 0;
    buttonDebounce[0] = digitalRead(A0) ? 20 : buttonDebounce[0] > 0 ? buttonDebounce[0] - mS : 0; // left
    buttonDebounce[1] = digitalRead(A1) ? 20 : buttonDebounce[1] > 0 ? buttonDebounce[1] - mS : 0; // top
    buttonDebounce[2] = digitalRead(A3) ? 20 : buttonDebounce[2] > 0 ? buttonDebounce[2] - mS : 0; // right
    buttonDebounce[3] = digitalRead(A4) ? 20 : buttonDebounce[3] > 0 ? buttonDebounce[3] - mS : 0; // bottom
    
    if (buttonDebounce[0] <= 0) // left
      axis[0] -= 512;
    if (buttonDebounce[1] <= 0) // top
      axis[1] -= 512;
    if (buttonDebounce[2] <= 0)
      axis[0] += 512;
    if (buttonDebounce[3] <= 0)
      axis[1] += 512;
          
  } else { // analog joystick option
    for (uint8_t i = 0; i < 2; i++) {
      int32_t v = 512L * (analogRead(A0 + i) - eeprom::gd.joyCentre[i]);
      if (v >= 0)
        axis[i] = v / (eeprom::gd.joyMax[i] - eeprom::gd.joyCentre[i]);
      else
        axis[i] = v / (eeprom::gd.joyCentre[i] - eeprom::gd.joyMin[i]);    
    }
  }
  
  button = oldButton;
  if (axis[0] > 256 && !(oldButton & 0x01)) button |= 0x01; // bit 0 is 'right push button' - comes on above 256
  if (axis[0] < 128 && (oldButton & 0x01)) button &= 0xFE; // goes back off below 128
  if (axis[0] < -256 && !(oldButton & 0x02)) button |= 0x02; // similarly bit 1 is 'left push button'
  if (axis[0] > -128 && (oldButton & 0x02)) button &= 0xFD;
  if (axis[1] > 256 && !(oldButton & 0x04)) button |= 0x04; // bit 2 is 'down push button'
  if (axis[1] < 128 && (oldButton & 0x04)) button &= 0x0B;
  if (axis[1] < -256 && !(oldButton & 0x08)) button |= 0x08; // bit 3 is 'up push button'
  if (axis[1] > -128 && (oldButton & 0x08)) button &= 0xF7;

  if (button)
    timeout = 30000; // reset timeout when a 'button' is pressed
  else
    increment = 1; // reset increment amount when no 'button' pressed

  if (!digitalRead(10)) {
    button |= 0x10;
    debounce = 100;
  }
  else if (debounce > 0) {
    debounce -= mS;
    button |= 0x10;
  }
  else
    button &= 0xEF;

  uint8_t t = button;
  
  if (button) {
    autoRepeat -= mS;
    if (autoRepeat > 0)
      button &= ~oldButton; // without auto-repeat bits only come on once for each activation
    else {
      button &= 0xEF; // pressing joystick down doesn't auto repeat but direction 'buttons' do
      autoRepeat += 333; // three auto-repeat keys per second
      increment *= 2;
      if (++increment > 50)
        increment = 50;
    }
  }
  else
    autoRepeat = 1000; // 1 second before first auto-repeat

  oldButton = t;  
}


