// remove // on following line to make #define active for rotary encoder version.  Leave // (#define then commented out) for analogue joystick/button joystick version
// #define ENCODER

#include <EEPROM.h>
#include <LiquidCrystal.h>
#include "eeprom.h"
#ifdef ENCODER
  #include "encoder.h"
#else
  #include "joystick.h"
#endif
#include "cppm_out.h"
#include "button.h"

float batteryVolts;
uint8_t menu; // which menu item is displayed (or 0 for 'main screens')
uint16_t remainingTime;
uint8_t mode; // motor running/stopped and joystick mode
uint8_t tempModelNum; // used when selecting model
int8_t posn; // pointer for editing different menus - reset to 0 when different menu selected
uint8_t buttonUp; // true if button is NOT currently pressed
uint8_t oldButtonState; // used when editing model name using button to go forward or back
uint8_t buttonMode; // decodes button presses and selects RIGHT, LEFT, KICK_UP, etc.
int16_t second; // one-second timer for buzzer and running countdown timer
boolean secondElapsed;

#ifdef ENCODER
  boolean encoderEdit; // true when editing a value, false when negotiating menus
  boolean encoderBeginEdit; // true on leading edge of edit (allows cursor to be switched on)
#endif

LiquidCrystal lcd(12, 11, 7, 6, 5, 4); // initialize library to use specified pins

void setup(void) {
  boolean fiveButton; // true indicates a digital top/bottom/left/right joystick instead of an analogue one
  
  pinMode(9, INPUT_PULLUP); // reset eeprom and calibrate joystick input.
  pinMode(2, INPUT_PULLUP); // main 'single channel' push button input
  pinMode(A3, INPUT_PULLUP); // for the five-button 'digital' joystick this is the 'right' button
  pinMode(A4, INPUT_PULLUP); // for the five-button 'digital' joystick this is the 'bottom' button  
  pinMode(8, OUTPUT); // buzzer
  digitalWrite(8, LOW); // buzzer off
  
  lcd.begin(16, 2);
  if (!digitalRead(9)) { // short pin 9 to ground to reset EEPROM at power on
    fiveButton = false; // default to an analogue joystick unless they press buttons on A3 or A4 during calibration
#ifndef ENCODER
    lcd.print(F("Cycle joystick."));
    lcd.setCursor(0, 1);
    lcd.print(F("Centre. Button."));
    eeprom::gd.joyMin[0] = 1023;
    eeprom::gd.joyMin[1] = 1023;
    eeprom::gd.joyMax[0] = 0;
    eeprom::gd.joyMax[1] = 0;

    while (digitalRead(2)) { // while main button not pressed sample joystick min max values (or buttons)
      int16_t a = analogRead(A0);
      if (a < eeprom::gd.joyMin[0])
        eeprom::gd.joyMin[0] = a;
      if (a > eeprom::gd.joyMax[0])
        eeprom::gd.joyMax[0] = a;
      
      a = analogRead(A1);
      if (a < eeprom::gd.joyMin[1])
        eeprom::gd.joyMin[1] = a;
      if (a > eeprom::gd.joyMax[1])
        eeprom::gd.joyMax[1] = a;

      if (!digitalRead(A3) || !digitalRead(A4)) // they must have built the digital joystick version
        fiveButton = true;
    }
    // when button is pressed sample centre values and reset eeprom
    eeprom::gd.joyCentre[0] = analogRead(A0);
    eeprom::gd.joyCentre[1] = analogRead(A1);
    eeprom::gd.fiveButton = fiveButton ? 1 : 0;
#endif

    eeprom::gd.modelNum = 1;
    eeprom::gd.batteryScale = 160;
    eeprom::gd.lowBatteryAlarm = 71;
    eeprom::putGD();
    
    eeprom::md.modelName[0] = 'U';
    eeprom::md.modelName[1] = 'n';
    eeprom::md.modelName[2] = 'u';
    eeprom::md.modelName[3] = 's';
    eeprom::md.modelName[4] = 'e';
    eeprom::md.modelName[5] = 'd';
    eeprom::md.modelName[6] = ' ';
    eeprom::md.modelName[7] = ' ';
    eeprom::md.modelName[8] = ' ';
    eeprom::md.modelName[9] = ' ';
    eeprom::md.modelName[10] = ' ';
    eeprom::md.modelName[11] = ' ';
    eeprom::md.modelName[12] = ' ';
    eeprom::md.modelName[13] = '\0';
    eeprom::md.rudderCentre = 1500;
    eeprom::md.rudderRight = 2000;
    eeprom::md.rudderLeft = 1000;
    eeprom::md.elevatorCentre = 1500;
    eeprom::md.elevatorUp = 2000;
    eeprom::md.elevatorDown = 1000;
    eeprom::md.motorOff = 1000;
    eeprom::md.motorOn = 1750;
    eeprom::md.motorSoftStart = 1000;
    eeprom::md.buttonSpeed = 250;
    eeprom::md.quickBlipEnabled = 1;
    eeprom::md.motorRunTime = 300;

    for (uint8_t i = 1; i <= 27; i++)
      eeprom::putMD(i);  
  }
  
  eeprom::getGD();
  fiveButton = (eeprom::gd.fiveButton != 0);
  eeprom::getMD(eeprom::gd.modelNum);
  cppm_out::set_uS(0, eeprom::md.rudderCentre); // channel 1 is rudder
  cppm_out::set_uS(1, eeprom::md.elevatorCentre); // channel 2 is elevator
  cppm_out::set_uS(2, eeprom::md.motorOff); // channel 3 is motor

#ifdef ENCODER
  encoder::start();
  encoderEdit = false;
  encoderBeginEdit = false;
#else
  joystick::start(fiveButton);
#endif
  
  button::start();
  pinMode(A2, INPUT); // battery monitor via 10K 22K divider to give 16-volt input range
  batteryVolts = (float)eeprom::gd.batteryScale * analogRead(A2) / 10240; // initialize as reading is smoothed
  cppm_out::start();
  menu = 0;
  mode = 0; 
  buttonMode = 0;
  second = 1000;
  
  // Serial.begin(115200);
}

void runTimer(boolean timerRunning) {
  if (timerRunning) {
    if (remainingTime) --remainingTime;
    if (remainingTime == 0) {
      if (mode < 3) // in normal button mode, stop motor
        mode = 0;
      else // in joystick mode, pause motor (but it can't be restarted when remainingTime is zero)
        mode = 4;
    }
  }
}

void displayTime(uint16_t seconds) {
#ifdef ENCODER
  if (menu == 1) // display time on bottom row when editing
    lcd.setCursor(11, 1);
  else
    lcd.setCursor(11, 0);
#else
  lcd.setCursor(11, 0);
#endif
  if (seconds < 600)
    lcd.print F(" ");
  
  lcd.print(seconds / 60);
  lcd.print F(":");

  seconds %= 60;
  if (seconds < 10)
    lcd.print F("0");
  
  lcd.print(seconds);
}

void displayTimer() {
  displayTime(remainingTime);
}

void topLine(const __FlashStringHelper *top) {
  lcd.setCursor(0, 0);
  lcd.print(top);  
}

void clearBottomLine() {
  lcd.setCursor(0, 1);
  lcd.print(F("                "));
}

void refresh(const __FlashStringHelper *top, const __FlashStringHelper *bottom, uint16_t mS) {
  static int16_t bottomLineCycler = 0;

  topLine(top);
  
  bottomLineCycler -= mS;
  if (bottomLineCycler <= 0)
    bottomLineCycler += 4500;

  lcd.setCursor(0, 1);
  if (bottomLineCycler >= 3000) { // display battery voltage
    lcd.print(F("Battery "));
    if (batteryVolts <= 0.1 * eeprom::gd.lowBatteryAlarm) {
      lcd.print(F("low"));
    }
    else {
      lcd.print(F("   "));
    }
    if (batteryVolts < 10.0)
      lcd.print(F(" "));

    lcd.print((int)batteryVolts);
    lcd.print(F("."));
    float f = (batteryVolts - (int)batteryVolts);
    lcd.print((int)(10 * f));
    lcd.print(F("V"));
  }
  else if (bottomLineCycler >= 1500) { // display model number and name
    if (eeprom::gd.modelNum < 10)
      lcd.print(F(" "));

    lcd.print(eeprom::gd.modelNum);
    lcd.print(F(" "));
    lcd.print(eeprom::md.modelName);
  }
  else // display current bottom line 'instruction'
    lcd.print(bottom);
}

boolean menuCheck() { // allow forward/back on encoder or up/down on joystick to cycle through main screen and menu items
#ifdef ENCODER
  if (menu && encoder::pressed) { // if encoder pressed when displaying menu items
    if (!encoderEdit) encoderBeginEdit = true;
    encoderEdit = true;
  }
  if (!encoderEdit) {
    int16_t change = 0;
    encoder::setMultiplier(1);
    encoderBeginEdit = false;
    change = encoder::change(change);
    if (change > 0) {
       clearBottomLine();
       change += menu;
       while (change > 17) change -= 18;
       
      if (change == 16)
        change++; // skip joystick mode choice when using encoder
      if (change >= 1 && change <= 4 && mode != 0)
        change = 5; // skip editing of motor time and model load/rename/copy if motor already armed

      menu = change;
    } else if (change < 0) {
      clearBottomLine();
      change += menu;
      while (change < 0) change += 18;

      if (change == 16)
        change--; // skip joystick mode choice when using encoder
      if (change >= 1 && change <= 4 && mode != 0)
        change = 0; // skip editing of motor time and model load/rename/copy if motor already armed

       menu = change;
    }
  }
  if (menu && !encoder::timeout) { // exit from menus back to main screens after thirty seconds of inactivity
    menu = 0;
    encoderEdit = false;
    encoder::setMultiplier(1);
    lcd.noBlink();
    return true;
  }
#else
  if (joystick::button & 0x04) { // down
    clearBottomLine();
    if (++menu > 17)
      menu = 0;
    else if (menu == 1)
      menu = 2; // skip motor run time on joystick version, as it's on the main screen
    if (menu >= 2  && menu <= 4  && mode != 0)
      menu = 5; // skip editing of model load/rename/copy if motor already armed
    return true;
  }
  if (joystick::button & 0x08) { // up
    clearBottomLine();
    if (!menu)
      menu = 17;
    else if (--menu == 1)
      menu = 0; // skip motor run time on joystick version, as it's on the main screen
    if (menu >= 2  && menu <= 4  && mode != 0)
      menu = 0; // skip editing of model load/rename/copy if motor already armed
    return true;
  }
  if (menu && joystick::button & 0x10) { // if joystick pressed when displaying menu items
    menu = 0; // return to main screens
    return true;
  }
  if (menu && !joystick::timeout) { // exit from menus back to main screens after thirty seconds of inactivity
    menu = 0;
    lcd.noBlink();
    return true;
  }
#endif
  return false;
}

void displayTempModel() {
  lcd.setCursor(0, 1);
  if (tempModelNum < 10)
    lcd.print(F(" "));

  lcd.print(tempModelNum);
  lcd.print(F(" "));
  lcd.print(eeprom::getModelName(tempModelNum));  
}

void selectTempModel() {
#ifdef ENCODER
  if (encoderEdit) {
    encoder::setMultiplier(1);
    int16_t change = 0;
    change = encoder::change(change);
    if (change || encoderBeginEdit) {
      tempModelNum += change;
      while (tempModelNum < 1) tempModelNum += 27;
      while (tempModelNum > 27) tempModelNum -= 27;
      lcd.noBlink();
      displayTempModel();
      lcd.setCursor(1, 1);
      lcd.blink();
    }
    if (!encoder::timeout) {
      lcd.noBlink();
      encoderEdit = false;
      encoder::setMultiplier(1);
      menu = 0;
      return;    
    }
  }
#else
  displayTempModel();

  if (joystick::button & 0x01) // right
    if (++tempModelNum > 27)
      tempModelNum = 1;
  if (joystick::button & 0x02) // left
    if (--tempModelNum < 1)
      tempModelNum = 27;
#endif
}

void edit_uS(const __FlashStringHelper *bottom, int16_t *uS)
{
#ifdef ENCODER
  int16_t change = 0;
  if (encoderEdit) {
    encoder::setMultiplier(10);    
    change = encoder::change(change);
    if (change) {
      *uS += change;
      if (*uS > 2200)
        *uS = 2200;
      if (*uS < 800)
        *uS = 800;
      eeprom::putMD(eeprom::gd.modelNum);
    }
    if (encoder::pressed) {
      encoderEdit = false;
    }
    if (!encoder::timeout) {
      lcd.noBlink();
      encoderEdit = false;
      encoder::setMultiplier(1);
      menu = 0;
      return;    
    }
  } else {
    lcd.noBlink();
    lcd.setCursor(0, 1);
    lcd.print(bottom); // should always be 9 print characters
    if (*uS < 1000)
      lcd.print(F(" "));
    lcd.print(*uS);
    lcd.print(F(" uS"));
    
    if (menuCheck()) {
      return;
    }
  }
#else
  if (menuCheck()) {
    return;
  }
  if (joystick::button & 0x01) { // right
    *uS += joystick::increment;
    if (*uS > 2200)
      *uS = 2200;
    eeprom::putMD(eeprom::gd.modelNum);
  }
  if (joystick::button & 0x02) { // left
    *uS -= joystick::increment;
    if (*uS < 800)
      *uS = 800;
    eeprom::putMD(eeprom::gd.modelNum);
  }
  if (!joystick::timeout) {
    menu = 0;
    return;    
  }

#endif

#ifdef ENCODER
  if (encoderEdit) {
    if (change) {
      lcd.noBlink();
      lcd.setCursor(0, 1);
      lcd.print(bottom); // should always be 9 print characters
      if (*uS < 1000)
        lcd.print(F(" "));
      lcd.print(*uS);
      lcd.print(F(" uS"));
    }
    lcd.setCursor(encoder::multiplying ? 11: 12, 1);
    lcd.blink();
  }
#else
  lcd.setCursor(0, 1);
  lcd.print(bottom); // should always be 9 print characters
  if (*uS < 1000)
    lcd.print(F(" "));
  lcd.print(*uS);
  lcd.print(F(" uS"));
#endif
}

void loop(void) {
  static uint32_t oldMillis = 0;
  uint32_t currentMillis = millis();
  uint16_t mS = currentMillis - oldMillis; // mS is milliseconds taken between consecutive calls to loop
  uint32_t buttonTime;

  if (mS > 100)
    mS = 0; // prevent startup delay causing problems
    
  oldMillis = currentMillis;

  if (mS) { // read battery voltage and smooth taking into account loop time
    batteryVolts *= 500 - mS;
    batteryVolts /= 500;
    batteryVolts += (float)eeprom::gd.batteryScale * analogRead(A2) * mS / 5120000;
  }

  if (second <= 500 && batteryVolts <= 0.1 * eeprom::gd.lowBatteryAlarm) {
    digitalWrite(8, HIGH); // buzzer on for battery warning
  }
  if (mode & 0x01 || mode == 2) { // if motor running, or paused in normal button mode - i.e. if timer is running
    if (remainingTime == 30 || remainingTime == 10 || remainingTime <= 5) { // 30-second, 10-second and last five second countdown bleeps
      digitalWrite(8, second > 500);
    }
    if (remainingTime == 1) { // last second - long bleep
      digitalWrite(8, HIGH);
    }
  }

  second -= mS;
  secondElapsed = false;
  if (second <= 0) {
    secondElapsed = true;
    second += 1000;
    digitalWrite(8, LOW); // buzzer off
  }

#ifdef ENCODER
  encoder::checkEncoder(mS);
  int16_t tens;
  int16_t change;
#else
  joystick::readJoy(mS);
#endif

  switch(menu) {
    case 0: // main screens
      tempModelNum = eeprom::gd.modelNum; // load copy of current model number for menu item 1
      switch (mode) {
        case 0: // motor not running - allow left-right editing of motor run time in 10s increments
          displayTime(eeprom::md.motorRunTime);
#ifdef ENCODER
          refresh(F("Motor time "), F("Press dial start"), mS);
          if (encoder::pressed) { // encoder pressed - start motor
            remainingTime = eeprom::md.motorRunTime;
            mode = 1; 
          }          
#else
          refresh(F("Motor time "), F("Press joy starts"), mS);
          if (joystick::button & 0x01) { // right
            eeprom::md.motorRunTime += 10;
            if (eeprom::md.motorRunTime > 5990)
              eeprom::md.motorRunTime = 5990; // limit max to 99:50
            eeprom::putMD(eeprom::gd.modelNum);
          }
          if (joystick::button & 0x02) { // left
            if (eeprom::md.motorRunTime > 10)
              eeprom::md.motorRunTime -= 10;
            eeprom::putMD(eeprom::gd.modelNum);
          }
          if (joystick::button & 0x10) { // joystick pressed - start motor
            remainingTime = eeprom::md.motorRunTime;
            mode = 1; 
          }
#endif
          menuCheck();
          break;
        case 1: // motor running (normal button control)
#ifdef ENCODER
          refresh(F("Remaining  "), F("Press dial stops"), mS);
#else
          refresh(F("Remaining  "), F("Press joy stops "), mS);
#endif
          runTimer(secondElapsed);
          displayTimer();
#ifdef ENCODER
          if (encoder::pressed) // encoder pressed - stop motor
            mode = 0;
#else
          if (joystick::button & 0x10) // joystick pressed - stop motor
            mode = 0;
#endif
          menuCheck();
          break;
        case 2: // motor paused (quick blip)
#ifdef ENCODER
          refresh(F("Paused     "), F("Press dial stops"), mS);
#else
          refresh(F("Paused     "), F("Press joy stops "), mS);
#endif
          // runTimer(secondElapsed);
          displayTimer();
#ifdef ENCODER
          if (encoder::pressed) // encoder pressed - stop motor
            mode = 0;
#else
          if (joystick::button & 0x10) // joystick pressed - stop motor
            mode = 0;
#endif
          menuCheck();
          break;
#ifndef ENCODER          
        case 3: // joystick mode, motor running
          refresh(F("=Joystick= "), F("Turn off to exit"), mS);
          runTimer(secondElapsed);
          displayTimer();
          break;
        case 4: // joystick mode, motor paused
          refresh(F("=Joystick= "), F("Turn off to exit"), mS);
          displayTimer();
          break;
#endif
      }
      break;

    case 1: // motor run time (encoder version only)
#ifdef ENCODER
      if (encoderEdit) {
        encoder::setMultiplier(1);
        tens = 0;
        tens = encoder::change(tens) * 10;
        if (tens || encoderBeginEdit) {
          eeprom::md.motorRunTime += tens;
          if (eeprom::md.motorRunTime > 5990)
            eeprom::md.motorRunTime = 5990; // limit max to 99:50
          if (eeprom::md.motorRunTime < 10)
            eeprom::md.motorRunTime = 10; // limit min to 00:10
          eeprom::putMD(eeprom::gd.modelNum);
          lcd.noBlink();
          displayTime(eeprom::md.motorRunTime);
          lcd.setCursor(14, 1);
          lcd.blink();        
        }
        encoderBeginEdit = false;
        if (encoder::pressed || !encoder::timeout) {
          encoderEdit = false;
          encoder::setMultiplier(1);
          lcd.noBlink();
          menu = 0;
        }
      } else {
        displayTime(eeprom::md.motorRunTime);
        topLine(F("Motor time      "));
        displayTime(eeprom::md.motorRunTime);
        menuCheck();
      }
#endif      
      break;
    case 2: // select model
#ifdef ENCODER
      if (encoderEdit) {
        if (encoderBeginEdit) {
          topLine(F("Press to use    "));
          encoderBeginEdit = false;
          lcd.setCursor(1, 1);
          lcd.blink();
        }
        if (encoder::pressed) {
          eeprom::gd.modelNum = tempModelNum;
          eeprom::putGD();
          eeprom::getMD(eeprom::gd.modelNum);
          encoderEdit = false;
          lcd.noBlink();          
          mode = 0;
          menu = 0;
        }
        if (!encoder::timeout) {
          encoderEdit = false;
          encoder::setMultiplier(1);
          lcd.noBlink();          
          menu = 0;          
        }
      } else {
        topLine(F("Press to select "));
        displayTempModel();
        menuCheck();
      }
      selectTempModel();
#else      
      topLine(F("Press to select "));
      selectTempModel();
      if (joystick::button & 0x10) { // pressed
        eeprom::gd.modelNum = tempModelNum;
        eeprom::putGD();
        eeprom::getMD(eeprom::gd.modelNum);
        mode = 0;
        menu = 0;
      }
      menuCheck();
#endif
      break;
    case 3: // model name
      topLine(F("Press: edit name"));
      displayTempModel();
#ifdef ENCODER
      if (encoder::pressed) {
        posn = 0;
        lcd.setCursor(3 + posn, 1);
        lcd.blink();
        oldButtonState = 0;
        menu = 255; // special menu for editing model name
        break;
      }
#else
      if (joystick::button & 0x10) { // pressed
        posn = 0;
        lcd.setCursor(3 + posn, 1);
        lcd.blink();
        menu = 255; // special menu for editing model name
        break;
      }
#endif
      menuCheck();
      break;
    case 4: // copy model to
#ifdef ENCODER
      if (encoderEdit) {
        if (encoderBeginEdit) {
          topLine(F("Press to copy to"));
          encoderBeginEdit = false;
          lcd.setCursor(1, 1);
          lcd.blink();
        }
        if (encoder::pressed) {
          eeprom::gd.modelNum = tempModelNum;
          eeprom::putGD();
          eeprom::putMD(eeprom::gd.modelNum);
          encoderEdit = false;
          lcd.noBlink();          
          mode = 0;
          menu = 0;
        }
        if (!encoder::timeout) {
          encoderEdit = false;
          encoder::setMultiplier(1);
          lcd.noBlink();          
          mode = 0;
          menu = 0;
        }
      } else {
        topLine(F("Copy model to   "));
        displayTempModel();
        menuCheck();
      }
      selectTempModel();
#else
      topLine(F("Copy model to   "));
      selectTempModel();
      if (joystick::button & 0x10) { // pressed
        eeprom::gd.modelNum = tempModelNum;
        eeprom::putGD();
        eeprom::putMD(eeprom::gd.modelNum);
        mode = 0;
        menu = 0;
      }
#endif
      menuCheck();
      break;
    case 5: // rudder centre pulse time
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Rudder centre   "));      
#else
      topLine(F("Rudder centre   "));
#endif
      edit_uS(F("         "), &(eeprom::md.rudderCentre));
      break;
    case 6: // rudder right pulse time
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Rudder 1-press  "));
#else
      topLine(F("Rudder 1-press  "));
#endif
      edit_uS(F("(Right)  "), &(eeprom::md.rudderRight));
      break;
    case 7: // rudder left pulse time
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Rudder 2-press  "));
#else
      topLine(F("Rudder 2-press  "));
#endif
      edit_uS(F("(Left)   "), &(eeprom::md.rudderLeft));
      break;
    case 8: // elevator centre pulse time
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Elevator centre "));
#else
      topLine(F("Elevator centre "));
#endif
      edit_uS(F("         "), &(eeprom::md.elevatorCentre));
      break;
    case 9: // elevator up pulse time
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Elevator up     "));
#else
      topLine(F("Elevator up     "));
#endif
      edit_uS(F("         "), &(eeprom::md.elevatorUp));
      break;
    case 10: // elevator down pulse time
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Elevator down   "));
#else
      topLine(F("Elevator down   "));
#endif
      edit_uS(F("         "), &(eeprom::md.elevatorDown));
      break;
    case 11: // motor off pulse time
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Motor off       "));
#else
      topLine(F("Motor off       "));
#endif
      edit_uS(F("         "), &(eeprom::md.motorOff));
      break;
    case 12: // motor off pulse time
#ifdef ENCODER
      if (!encoderEdit)
        topLine(F("Motor on        "));
#else
      topLine(F("Motor on        "));
#endif
      edit_uS(F("         "), &(eeprom::md.motorOn));
      break;
    case 13: // motor soft start time (milliseconds)
#ifdef ENCODER
      if (encoderEdit) {
        change = 0;
        change = encoder::change(change);
        if (change || encoderBeginEdit) {
          encoderBeginEdit = false;
          eeprom::md.motorSoftStart += change;          
          if (eeprom::md.motorSoftStart > 5000)
            eeprom::md.motorSoftStart = 5000;
          if (eeprom::md.motorSoftStart < 0)
            eeprom::md.motorSoftStart = 0;

          eeprom::putMD(eeprom::gd.modelNum);
          lcd.noBlink();
          lcd.setCursor(0, 1);
          lcd.print(F("         ")); // should always be 9 print characters
          if (eeprom::md.motorSoftStart < 1000)
            lcd.print(F(" "));
          if (eeprom::md.motorSoftStart < 100)
            lcd.print(F(" "));
          if (eeprom::md.motorSoftStart < 10)
            lcd.print(F(" "));
    
          lcd.print(eeprom::md.motorSoftStart);
          lcd.print(F(" mS"));
          lcd.setCursor(encoder::multiplying ? 11: 12, 1);
          lcd.blink();        
        }
        if (encoder::pressed) {
          lcd.noBlink();
          encoderEdit = false;
        }
        if (!encoder::timeout) {
          lcd.noBlink();
          encoderEdit = false;
          encoder::setMultiplier(1);
          menu = 0;
          return;    
        }
      } else {
        topLine(F("Motor soft start"));
        encoder::setMultiplier(10);
        if (menuCheck())
          break;

        lcd.setCursor(0, 1);
        lcd.print(F("         ")); // should always be 9 print characters
        if (eeprom::md.motorSoftStart < 1000)
          lcd.print(F(" "));
        if (eeprom::md.motorSoftStart < 100)
          lcd.print(F(" "));
        if (eeprom::md.motorSoftStart < 10)
          lcd.print(F(" "));
  
        lcd.print(eeprom::md.motorSoftStart);
        lcd.print(F(" mS"));        
      }
#else
      topLine(F("Motor soft start"));
      if (menuCheck())
        break;
      if (joystick::button & 0x01) { // right
        eeprom::md.motorSoftStart += joystick::increment;
        if (eeprom::md.motorSoftStart > 5000)
          eeprom::md.motorSoftStart = 5000;
        eeprom::putMD(eeprom::gd.modelNum);
      }
      if (joystick::button & 0x02) { // left
        if (joystick::increment >= eeprom::md.motorSoftStart)
          eeprom::md.motorSoftStart = 0;
        else
          eeprom::md.motorSoftStart -= joystick::increment;

        eeprom::putMD(eeprom::gd.modelNum);
      }
      lcd.setCursor(0, 1);
      lcd.print(F("         ")); // should always be 9 print characters
      if (eeprom::md.motorSoftStart < 1000)
        lcd.print(F(" "));
      if (eeprom::md.motorSoftStart < 100)
        lcd.print(F(" "));
      if (eeprom::md.motorSoftStart < 10)
        lcd.print(F(" "));

      lcd.print(eeprom::md.motorSoftStart);
      lcd.print(F(" mS"));
#endif
      break;      
    case 14: // button speed (milliseconds)
#ifdef ENCODER
      if (encoderEdit) {
        change = 0;
        change = encoder::change(change);
        if (change || encoderBeginEdit) {
          encoderBeginEdit = false;
          eeprom::md.buttonSpeed += change;          
          if (eeprom::md.buttonSpeed > 1000)
            eeprom::md.buttonSpeed = 1000;
          if (eeprom::md.buttonSpeed < 100)
            eeprom::md.buttonSpeed = 100;

          eeprom::putMD(eeprom::gd.modelNum);
          lcd.noBlink();
          lcd.setCursor(0, 1);
          lcd.print(F("         ")); // should always be 9 print characters
          if (eeprom::md.buttonSpeed < 1000)
            lcd.print(F(" "));
    
          lcd.print(eeprom::md.buttonSpeed);
          lcd.print(F(" mS"));          
          lcd.setCursor(encoder::multiplying ? 11: 12, 1);
          lcd.blink();        
        }
        if (encoder::pressed) {
          lcd.noBlink();
          encoderEdit = false;
        }
        if (!encoder::timeout) {
          lcd.noBlink();
          encoderEdit = false;
          encoder::setMultiplier(1);
          menu = 0;
          return;    
        }
      } else {
        topLine(F("Button speed    "));
        encoder::setMultiplier(10);
        if (menuCheck())
          break;

        lcd.setCursor(0, 1);
        lcd.print(F("         ")); // should always be 9 print characters
        if (eeprom::md.buttonSpeed < 1000)
          lcd.print(F(" "));
  
        lcd.print(eeprom::md.buttonSpeed);
        lcd.print(F(" mS"));        
      }
#else
      topLine(F("Button speed    "));
      if (menuCheck())
        break;
      if (joystick::button & 0x01) { // right
        eeprom::md.buttonSpeed += joystick::increment;
        if (eeprom::md.buttonSpeed > 1000)
          eeprom::md.buttonSpeed = 1000;
        eeprom::putMD(eeprom::gd.modelNum);
      }
      if (joystick::button & 0x02) { // left
        if (joystick::increment >= eeprom::md.buttonSpeed)
          eeprom::md.buttonSpeed = 100;
        else {
          eeprom::md.buttonSpeed -= joystick::increment;
          if (eeprom::md.buttonSpeed < 100)
            eeprom::md.buttonSpeed = 100;
        }
        eeprom::putMD(eeprom::gd.modelNum);
      }
      lcd.setCursor(0, 1);
      lcd.print(F("         ")); // should always be 9 print characters
      if (eeprom::md.buttonSpeed < 1000)
        lcd.print(F(" "));

      lcd.print(eeprom::md.buttonSpeed);
      lcd.print(F(" mS"));
#endif
      break;
    case 15: // motor control (quick blip) enable
      topLine(F("Motor control by"));
      if (menuCheck())
        break;
#ifdef ENCODER
      if (encoder::pressed) {
        if (eeprom::md.quickBlipEnabled)
          eeprom::md.quickBlipEnabled = 0;
        else
          eeprom::md.quickBlipEnabled = 1;
        eeprom::putMD(eeprom::gd.modelNum);
      }
      encoderEdit = false;
#else        
      if (joystick::button & 0x03) { // left or right toggles true/false
        if (eeprom::md.quickBlipEnabled)
          eeprom::md.quickBlipEnabled = 0;
        else
          eeprom::md.quickBlipEnabled = 1;
        eeprom::putMD(eeprom::gd.modelNum);
      }
#endif
      lcd.setCursor(0, 1);
      lcd.print(F("quick blip?  "));
      if (eeprom::md.quickBlipEnabled)
        lcd.print(F("Yes"));
      else
        lcd.print(F("No "));
      
      break;
#ifndef ENCODER
    case 16: // press for joystick mode
      topLine(F("Press joy starts"));
      lcd.setCursor(0, 1);
      lcd.print(F("joystick mode   "));
      
      if (joystick::button & 0x10) { // pressed
        remainingTime = eeprom::md.motorRunTime;        
        mode = 4;
        menu = 0;
      }
      menuCheck();
      break;
#endif
    case 17: // low battery alarm
#ifdef ENCODER
      if (encoderEdit) {
        change = 0;
        change = encoder::change(change);
        if (change || encoderBeginEdit) {
          encoderBeginEdit = false;
          eeprom::gd.lowBatteryAlarm += change;          
          if (eeprom::gd.lowBatteryAlarm > 160)
            eeprom::gd.lowBatteryAlarm = 160;
          if (eeprom::gd.lowBatteryAlarm < 65)
            eeprom::gd.lowBatteryAlarm = 65;

          eeprom::putGD();
          lcd.noBlink();
          lcd.setCursor(0, 1);
          lcd.print(F("alarm      "));
          if (eeprom::gd.lowBatteryAlarm < 100)
            lcd.print(F(" "));
    
          lcd.print(eeprom::gd.lowBatteryAlarm / 10);
          lcd.print(F("."));
          lcd.print(eeprom::gd.lowBatteryAlarm % 10);
          lcd.print(F("V"));
          lcd.setCursor(14, 1);
          lcd.blink();        
        }
        if (encoder::pressed) {
          lcd.noBlink();
          encoderEdit = false;
        }
        if (!encoder::timeout) {
          lcd.noBlink();
          encoderEdit = false;
          encoder::setMultiplier(1);
          menu = 0;
          return;    
        }
      } else {
        topLine(F("Low battery     "));
        encoder::setMultiplier(1);
        if (menuCheck())
          break;

        lcd.setCursor(0, 1);
        lcd.print(F("alarm      "));
        if (eeprom::gd.lowBatteryAlarm < 100)
          lcd.print(F(" "));
  
        lcd.print(eeprom::gd.lowBatteryAlarm / 10);
        lcd.print(F("."));
        lcd.print(eeprom::gd.lowBatteryAlarm % 10);
        lcd.print(F("V"));
      }
#else
      topLine(F("Low battery     "));
      if (menuCheck())
        break;
      if (joystick::button & 0x01) { // right
        eeprom::gd.lowBatteryAlarm += joystick::increment;
        if (eeprom::gd.lowBatteryAlarm > 160)
          eeprom::gd.lowBatteryAlarm = 160;
        eeprom::putGD();
      }
      if (joystick::button & 0x02) { // left
        if (joystick::increment >= eeprom::gd.lowBatteryAlarm)
          eeprom::gd.lowBatteryAlarm = 65;
        else {
          eeprom::gd.lowBatteryAlarm -= joystick::increment;
          if (eeprom::gd.lowBatteryAlarm < 65)
            eeprom::gd.lowBatteryAlarm = 65;
        }
        eeprom::putGD();
      }
      lcd.setCursor(0, 1);
      lcd.print(F("alarm      "));
      if (eeprom::gd.lowBatteryAlarm < 100)
        lcd.print(F(" "));

      lcd.print(eeprom::gd.lowBatteryAlarm / 10);
      lcd.print(F("."));
      lcd.print(eeprom::gd.lowBatteryAlarm % 10);
      lcd.print(F("V"));
#endif
      break;      
    case 255: // editing model name
#ifdef ENCODER
    change = 0;
    change = encoder::change(change);
    buttonTime = button::milliseconds(&buttonUp);
    if (buttonUp) oldButtonState = 0;
    if (!buttonUp && oldButtonState == 0 && buttonTime > 20) { // real (non contact bounce) button press advances one character
      oldButtonState = 1;
      if (++posn > 12)
          posn = 0;
      change = 0;
    }
    if (!buttonUp && oldButtonState == 2 && buttonTime > 1000) { // long button press steps back one character
      oldButtonState = 3;
      posn -= 2;
      if (posn < 0)
          posn += 13;
      change = 0;
    }
    if (change) {
      char c = eeprom::modelName[posn] + change;
        while (c < 32)
          c += 95;
        while (c > 126)
          c -= 95;
      eeprom::modelName[posn] = c;
    }
    if (change || oldButtonState == 1 || oldButtonState == 3) {
      lcd.noBlink();
      lcd.setCursor(3, 1);
      lcd.print(eeprom::modelName);
      lcd.setCursor(3 + posn, 1);
      lcd.blink();
      if (oldButtonState == 1 || oldButtonState == 3) oldButtonState++;            
    }
    if (encoder::pressed) {
      for (posn = 0; posn < 13; posn++)
        eeprom::md.modelName[posn] = eeprom::modelName[posn];
      eeprom::putMD(eeprom::gd.modelNum);
      menu = 0;
      lcd.noBlink();
    }
    
    if (!encoder::timeout) { // inactivity timeout back to main screens (no save)
      menu = 0;
      lcd.noBlink();
      encoderEdit = false;
      encoder::setMultiplier(1);      
    }
#else    
      if (joystick::button & 0x01) // right
        if (++posn > 12)
          posn = 0;
      if (joystick::button & 0x02) // left
        if (!posn)
          posn = 12;
        else
          --posn;
      if (joystick::button & 0x04) { // down
        char c = eeprom::modelName[posn] - 1;
        if (c < 32)
          c = 126;
        eeprom::modelName[posn] = c;
      }
      if (joystick::button & 0x08) { // up
        char c = eeprom::modelName[posn] + 1;
        if (c > 126)
          c = 32;
        eeprom::modelName[posn] = c;
      }

      if (joystick::button) { // after any button press refresh
        lcd.noBlink();
        lcd.setCursor(3, 1);
        lcd.print(eeprom::modelName);
        lcd.setCursor(3 + posn, 1);
        lcd.blink();
      }
      
      if (joystick::button & 0x10) { // pressed
        for (posn = 0; posn < 13; posn++)
          eeprom::md.modelName[posn] = eeprom::modelName[posn];
        eeprom::putMD(eeprom::gd.modelNum);
        menu = 0;
        lcd.noBlink();
      }
      
      if (!joystick::timeout) { // inactivity timeout back to main screens (no save)
        menu = 0;
        lcd.noBlink();
      }
#endif
      break;
    default:
      lcd.setCursor(0, 1);
      lcd.print(menu);
      lcd.print(F(" = menu  "));
      break;
  }

  // cppm_out::set_uS(0, eeprom::md.rudderCentre);

  buttonTime = button::milliseconds(&buttonUp);

  if (mode == 3 || mode == 4) { // joystick mode (3 = motor running, 4 = motor paused)  button toggles motor
    if (buttonUp) {
      buttonMode = 0;
    } else if (!buttonMode && buttonTime > 20) { // each real (non contact bouncing) press toggles motor on/off
       buttonMode = 1;
       mode = 7 - mode;
      if (!remainingTime) {
        mode = 4; // can't restart motor once timer has run out
      }
    }
    // work rudder from joystick
#ifndef ENCODER      
    if (joystick::axis[0] > 20) { // right
      int32_t posn = (int32_t)eeprom::md.rudderRight - (int32_t)eeprom::md.rudderCentre;
      posn *= (int32_t)joystick::axis[0];
      posn /= 512L;
      posn += (int32_t)eeprom::md.rudderCentre;
      cppm_out::set_uS(0, (int16_t)posn);
    } else if(joystick::axis[0] < -20) { // left
      int32_t posn = (int32_t)eeprom::md.rudderLeft - (int32_t)eeprom::md.rudderCentre;
      posn *= (int32_t)joystick::axis[0];
      posn /= -512L;
      posn += (int32_t)eeprom::md.rudderCentre;
      cppm_out::set_uS(0, (int16_t)posn);
    } else { // rudder centre
      cppm_out::set_uS(0, eeprom::md.rudderCentre);
    }

    // work elevator from joystick
    if (joystick::axis[1] > 20) { // joystick down -> up elevator
      int32_t posn = (int32_t)eeprom::md.elevatorUp - (int32_t)eeprom::md.elevatorCentre;
      posn *= (int32_t)joystick::axis[1];
      posn /= 512L;
      posn += (int32_t)eeprom::md.elevatorCentre;
      cppm_out::set_uS(1, (int16_t)posn);
    } else if (joystick::axis[1] < -20) { // joystick up -> down elevator
      int32_t posn = (int32_t)eeprom::md.elevatorDown - (int32_t)eeprom::md.elevatorCentre;
      posn *= (int32_t)joystick::axis[1];
      posn /= -512L;
      posn += (int32_t)eeprom::md.elevatorCentre;
      cppm_out::set_uS(1, (int16_t)posn);
    } else { // elevator centre
      cppm_out::set_uS(1, eeprom::md.elevatorCentre);
    }
#endif
  } else { // non joystick mode
    if (buttonUp) {
      if (buttonMode && buttonTime > eeprom::md.buttonSpeed * 2) { // when button released for long enough, back to idle mode (controls centred)
          buttonMode = 0;
      } else if (buttonTime > 20) { // a real release - not switch bounce
        switch (buttonMode) {
          case 1: // previously 'rotation started' - this is a quick blip so toggle motor state if enabled and return to idle
            if (eeprom::md.quickBlipEnabled) {
              if (mode == 1 || mode == 2) {
                mode = 3 - mode; // toggle motor state
                buttonMode = 0; // back to idle
              }
            } else {
              buttonMode = 4; // quick blip not enabled, so treat as a normal button press
            }
            break;
          case 2: // longer than a quick blip, but not 'RIGHT' 
            buttonMode = 4;
            break;
          case 3: // previously RIGHT
          case 6: // previously LEFT
          case 8: // previously KICK_UP
            buttonMode = 0; // back to idle
            break;
          case 5: // second release, (not 'LEFT')
            buttonMode = 7;
            break;
        }
      }
    } else if (buttonTime > 20) { // button really down - not switch bounce
      switch (buttonMode) {
        case 0: // idle - move to 'rotation started'
          buttonMode = 1;
          break;
        case 4:
          buttonMode = 5;
          break;
      }
      if (buttonTime > eeprom::md.buttonSpeed / 2) { // longer than a quick blip
        switch (buttonMode) {
          case 1:
            buttonMode = 2; // move to 'rotated past quick blip
            break;
        }
      }
      if (buttonTime > eeprom::md.buttonSpeed) { // button continued to be down long enough to move either rudder or elevator
        switch (buttonMode) {
          case 2: // single press and hold - RIGHT
            buttonMode = 3;
            break;
          case 5: // press, release, press and hold - LEFT
            buttonMode = 6;
            break;
          case 7: // press, release, press, release, press and hold - KICK UP
            buttonMode = 8;
            break;  
        }
      }
    }

    switch (buttonMode) {
      case 3:
        cppm_out::set_uS(0, eeprom::md.rudderRight);
        break;
      case 6:
        cppm_out::set_uS(0, eeprom::md.rudderLeft);
        break;
      case 8:
        cppm_out::set_uS(1, eeprom::md.elevatorUp);
        break;
      default:
        cppm_out::set_uS(0, eeprom::md.rudderCentre);
        cppm_out::set_uS(1, eeprom::md.elevatorCentre);
      break;
    }
  }
  
  if (mode & 0x01) { // motor is running
    if (mS >= eeprom::md.motorSoftStart) {
      cppm_out::set_uS(2, eeprom::md.motorOn);
    } else {
      int16_t motorSpeed;
      if (eeprom::md.motorOn >= eeprom::md.motorOff) {
        motorSpeed = (eeprom::md.motorOn - eeprom::md.motorOff) * mS / eeprom::md.motorSoftStart; // change in speed this loop for soft start
        motorSpeed += cppm_out::get_uS(2);
      } else {
        motorSpeed = (eeprom::md.motorOff - eeprom::md.motorOn) * mS / eeprom::md.motorSoftStart; // change in speed this loop for soft start      
        motorSpeed = cppm_out::get_uS(2) - motorSpeed;
      }
      if (eeprom::md.motorOn > eeprom::md.motorOff && motorSpeed > eeprom::md.motorOn) motorSpeed = eeprom::md.motorOn;
      if (eeprom::md.motorOn < eeprom::md.motorOff && motorSpeed < eeprom::md.motorOn) motorSpeed = eeprom::md.motorOn;
      cppm_out::set_uS(2, motorSpeed);
    }
  } else { // motor stopped
    cppm_out::set_uS(2, eeprom::md.motorOff);    
  }

  /*
  Serial.print(menu);
  Serial.print("\t");
  Serial.print(mode);
  Serial.print("\t");
  Serial.print(encoder::timeout);
  Serial.print("\t");
  Serial.print(encoder::turningTime);
  Serial.print("\t");
  Serial.println(encoder::multiplying);

  Serial.print(buttonUp);
  Serial.print("\t");
  Serial.print(buttonTime);
  Serial.print("\t");
  Serial.print(mS);
  Serial.print("\t");
  Serial.print(menu);
  Serial.print("\t");
  Serial.print(mode);
  Serial.print("\t");
  Serial.print(joystick::axis[0]);
  Serial.print("\t");
  Serial.print(joystick::axis[1]);
  Serial.print("\t");
  Serial.println(buttonMode);
  */
}
